# Based on EVA, BEIT, timm and DeiT code bases
# https://github.com/baaivision/EVA
# https://github.com/rwightman/pytorch-image-models/tree/master/timm
# https://github.com/microsoft/unilm/tree/master/beit
# https://github.com/facebookresearch/deit/
# https://github.com/facebookresearch/dino
# --------------------------------------------------------'
import math
from functools import partial

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.checkpoint as checkpoint
from timm.layers import drop_path, to_2tuple, trunc_normal_
from timm.models import register_model

from ..common.dist_utils import download_cached_file

def _cfg(url='', **kwargs):
    return {
        'url': url,
        'num_classes': 1000, 'input_size': (3, 224, 224), 'pool_size': None,
        'crop_pct': .9, 'interpolation': 'bicubic',
        'mean': (0.5, 0.5, 0.5), 'std': (0.5, 0.5, 0.5),
        **kwargs
    }


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks).
    """
    def __init__(self, drop_prob=None):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training)
    
    def extra_repr(self) -> str:
        return 'p={}'.format(self.drop_prob)


class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        # x = self.drop(x)
        # commit this for the orignal BERT implement 
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Attention(nn.Module):
    def __init__(
            self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0.,
            proj_drop=0., window_size=None, attn_head_dim=None):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        if attn_head_dim is not None:
            head_dim = attn_head_dim
        all_head_dim = head_dim * self.num_heads
        self.scale = qk_scale or head_dim ** -0.5

        self.qkv = nn.Linear(dim, all_head_dim * 3, bias=False)
        if qkv_bias:
            self.q_bias = nn.Parameter(torch.zeros(all_head_dim))
            self.v_bias = nn.Parameter(torch.zeros(all_head_dim))
        else:
            self.q_bias = None
            self.v_bias = None

        if window_size:
            self.window_size = window_size
            self.num_relative_distance = (2 * window_size[0] - 1) * (2 * window_size[1] - 1) + 3
            self.relative_position_bias_table = nn.Parameter(
                torch.zeros(self.num_relative_distance, num_heads))  # 2*Wh-1 * 2*Ww-1, nH
            # cls to token & token 2 cls & cls to cls

            # get pair-wise relative position index for each token inside the window
            coords_h = torch.arange(window_size[0])
            coords_w = torch.arange(window_size[1])
            coords = torch.stack(torch.meshgrid([coords_h, coords_w]))  # 2, Wh, Ww
            coords_flatten = torch.flatten(coords, 1)  # 2, Wh*Ww
            relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]  # 2, Wh*Ww, Wh*Ww
            relative_coords = relative_coords.permute(1, 2, 0).contiguous()  # Wh*Ww, Wh*Ww, 2
            relative_coords[:, :, 0] += window_size[0] - 1  # shift to start from 0
            relative_coords[:, :, 1] += window_size[1] - 1
            relative_coords[:, :, 0] *= 2 * window_size[1] - 1
            relative_position_index = \
                torch.zeros(size=(window_size[0] * window_size[1] + 1, ) * 2, dtype=relative_coords.dtype)
            relative_position_index[1:, 1:] = relative_coords.sum(-1)  # Wh*Ww, Wh*Ww
            relative_position_index[0, 0:] = self.num_relative_distance - 3
            relative_position_index[0:, 0] = self.num_relative_distance - 2
            relative_position_index[0, 0] = self.num_relative_distance - 1

            self.register_buffer("relative_position_index", relative_position_index)
        else:
            self.window_size = None
            self.relative_position_bias_table = None
            self.relative_position_index = None

        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(all_head_dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x, rel_pos_bias=None):
        B, N, C = x.shape
        qkv_bias = None
        if self.q_bias is not None:
            qkv_bias = torch.cat((self.q_bias, torch.zeros_like(self.v_bias, requires_grad=False), self.v_bias))
        # qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        qkv = F.linear(input=x, weight=self.qkv.weight, bias=qkv_bias)
        qkv = qkv.reshape(B, N, 3, self.num_heads, -1).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]   # make torchscript happy (cannot use tensor as tuple)

        q = q * self.scale
        attn = (q @ k.transpose(-2, -1))

        if self.relative_position_bias_table is not None:
            relative_position_bias = \
                self.relative_position_bias_table[self.relative_position_index.view(-1)].view(
                    self.window_size[0] * self.window_size[1] + 1,
                    self.window_size[0] * self.window_size[1] + 1, -1)  # Wh*Ww,Wh*Ww,nH
            relative_position_bias = relative_position_bias.permute(2, 0, 1).contiguous()  # nH, Wh*Ww, Wh*Ww
            attn = attn + relative_position_bias.unsqueeze(0)

        if rel_pos_bias is not None:
            attn = attn + rel_pos_bias
        
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(B, N, -1)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class Block(nn.Module):

    def __init__(self, dim, num_heads, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop=0., attn_drop=0.,
                 drop_path=0., init_values=None, act_layer=nn.GELU, norm_layer=nn.LayerNorm,
                 window_size=None, attn_head_dim=None):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim, num_heads=num_heads, qkv_bias=qkv_bias, qk_scale=qk_scale,
            attn_drop=attn_drop, proj_drop=drop, window_size=window_size, attn_head_dim=attn_head_dim)
        # NOTE: drop path for stochastic depth, we shall see if this is better than dropout here
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

        if init_values is not None and init_values > 0:
            self.gamma_1 = nn.Parameter(init_values * torch.ones((dim)),requires_grad=True)
            self.gamma_2 = nn.Parameter(init_values * torch.ones((dim)),requires_grad=True)
        else:
            self.gamma_1, self.gamma_2 = None, None

    def forward(self, x, rel_pos_bias=None):
        if self.gamma_1 is None:
            x = x + self.drop_path(self.attn(self.norm1(x), rel_pos_bias=rel_pos_bias))
            x = x + self.drop_path(self.mlp(self.norm2(x)))
        else:
            x = x + self.drop_path(self.gamma_1 * self.attn(self.norm1(x), rel_pos_bias=rel_pos_bias))
            x = x + self.drop_path(self.gamma_2 * self.mlp(self.norm2(x)))
        return x


class PatchEmbed(nn.Module):
    """ Image to Patch Embedding
    """
    def __init__(self, img_size=224, patch_size=16, in_chans=3, embed_dim=768):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        num_patches = (img_size[1] // patch_size[1]) * (img_size[0] // patch_size[0])
        self.patch_shape = (img_size[0] // patch_size[0], img_size[1] // patch_size[1])
        self.img_size = img_size
        self.patch_size = patch_size
        self.num_patches = num_patches

        self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_size, stride=patch_size)

    def forward(self, x, **kwargs):
        B, C, H, W = x.shape
        # FIXME look at relaxing size constraints
        assert H == self.img_size[0] and W == self.img_size[1], \
            f"Input image size ({H}*{W}) doesn't match model ({self.img_size[0]}*{self.img_size[1]})."
        x = self.proj(x).flatten(2).transpose(1, 2)
        return x


class RelativePositionBias(nn.Module):

    def __init__(self, window_size, num_heads):
        super().__init__()
        self.window_size = window_size
        self.num_relative_distance = (2 * window_size[0] - 1) * (2 * window_size[1] - 1) + 3
        self.relative_position_bias_table = nn.Parameter(
            torch.zeros(self.num_relative_distance, num_heads))  # 2*Wh-1 * 2*Ww-1, nH
        # cls to token & token 2 cls & cls to cls

        # get pair-wise relative position index for each token inside the window
        coords_h = torch.arange(window_size[0])
        coords_w = torch.arange(window_size[1])
        coords = torch.stack(torch.meshgrid([coords_h, coords_w]))  # 2, Wh, Ww
        coords_flatten = torch.flatten(coords, 1)  # 2, Wh*Ww
        relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]  # 2, Wh*Ww, Wh*Ww
        relative_coords = relative_coords.permute(1, 2, 0).contiguous()  # Wh*Ww, Wh*Ww, 2
        relative_coords[:, :, 0] += window_size[0] - 1  # shift to start from 0
        relative_coords[:, :, 1] += window_size[1] - 1
        relative_coords[:, :, 0] *= 2 * window_size[1] - 1
        relative_position_index = \
            torch.zeros(size=(window_size[0] * window_size[1] + 1,) * 2, dtype=relative_coords.dtype)
        relative_position_index[1:, 1:] = relative_coords.sum(-1)  # Wh*Ww, Wh*Ww
        relative_position_index[0, 0:] = self.num_relative_distance - 3
        relative_position_index[0:, 0] = self.num_relative_distance - 2
        relative_position_index[0, 0] = self.num_relative_distance - 1

        self.register_buffer("relative_position_index", relative_position_index)

        # trunc_normal_(self.relative_position_bias_table, std=.02)

    def forward(self):
        relative_position_bias = \
            self.relative_position_bias_table[self.relative_position_index.view(-1)].view(
                self.window_size[0] * self.window_size[1] + 1,
                self.window_size[0] * self.window_size[1] + 1, -1)  # Wh*Ww,Wh*Ww,nH
        return relative_position_bias.permute(2, 0, 1).contiguous()  # nH, Wh*Ww, Wh*Ww


class VisionTransformer(nn.Module):
    """ Vision Transformer with support for patch or hybrid CNN input stage
    """
    def __init__(self, img_size=224, patch_size=16, in_chans=3, num_classes=1000, embed_dim=768, depth=12,
                 num_heads=12, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop_rate=0., attn_drop_rate=0.,
                 drop_path_rate=0., norm_layer=nn.LayerNorm, init_values=None,
                 use_abs_pos_emb=True, use_rel_pos_bias=False, use_shared_rel_pos_bias=False,
                 use_mean_pooling=True, init_scale=0.001, use_checkpoint=False):
        super().__init__()
        self.image_size = img_size
        self.num_classes = num_classes
        self.num_features = self.embed_dim = embed_dim  # num_features for consistency with other models

        self.patch_embed = PatchEmbed(
            img_size=img_size, patch_size=patch_size, in_chans=in_chans, embed_dim=embed_dim)
        num_patches = self.patch_embed.num_patches

        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        if use_abs_pos_emb:
            self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + 1, embed_dim))
        else:
            self.pos_embed = None
        self.pos_drop = nn.Dropout(p=drop_rate)

        if use_shared_rel_pos_bias:
            self.rel_pos_bias = RelativePositionBias(window_size=self.patch_embed.patch_shape, num_heads=num_heads)
        else:
            self.rel_pos_bias = None
        self.use_checkpoint = use_checkpoint
        
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  # stochastic depth decay rule
        self.use_rel_pos_bias = use_rel_pos_bias
        self.blocks = nn.ModuleList([
            Block(
                dim=embed_dim, num_heads=num_heads, mlp_ratio=mlp_ratio, qkv_bias=qkv_bias, qk_scale=qk_scale,
                drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[i], norm_layer=norm_layer,
                init_values=init_values, window_size=self.patch_embed.patch_shape if use_rel_pos_bias else None)
            for i in range(depth)])
#         self.norm = nn.Identity() if use_mean_pooling else norm_layer(embed_dim)
#         self.fc_norm = norm_layer(embed_dim) if use_mean_pooling else None
#         self.head = nn.Linear(embed_dim, num_classes) if num_classes > 0 else nn.Identity()

        if self.pos_embed is not None:
            trunc_normal_(self.pos_embed, std=.02)
        trunc_normal_(self.cls_token, std=.02)
        # trunc_normal_(self.mask_token, std=.02)
#         if isinstance(self.head, nn.Linear):
#             trunc_normal_(self.head.weight, std=.02)
        self.apply(self._init_weights)
        self.fix_init_weight()
#         if isinstance(self.head, nn.Linear):
#             self.head.weight.data.mul_(init_scale)
#             self.head.bias.data.mul_(init_scale)

    def fix_init_weight(self):
        def rescale(param, layer_id):
            param.div_(math.sqrt(2.0 * layer_id))

        for layer_id, layer in enumerate(self.blocks):
            rescale(layer.attn.proj.weight.data, layer_id + 1)
            rescale(layer.mlp.fc2.weight.data, layer_id + 1)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def get_classifier(self):
        return self.head

    def reset_classifier(self, num_classes, global_pool=''):
        self.num_classes = num_classes
        self.head = nn.Linear(self.embed_dim, num_classes) if num_classes > 0 else nn.Identity()

    def forward_features(self, x):
        x = self.patch_embed(x)
        batch_size, seq_len, _ = x.size()

        cls_tokens = self.cls_token.expand(batch_size, -1, -1)  # stole cls_tokens impl from Phil Wang, thanks
        x = torch.cat((cls_tokens, x), dim=1)
        if self.pos_embed is not None:
            x = x + self.pos_embed
        x = self.pos_drop(x)

        rel_pos_bias = self.rel_pos_bias() if self.rel_pos_bias is not None else None
        for blk in self.blocks:
            if self.use_checkpoint:
                x = checkpoint.checkpoint(blk, x, rel_pos_bias)
            else:
                x = blk(x, rel_pos_bias)
        return x
#         x = self.norm(x)

#         if self.fc_norm is not None:
#             t = x[:, 1:, :]
#             return self.fc_norm(t.mean(1))
#         else:
#             return x[:, 0]

    def forward(self, x):
        x = self.forward_features(x)
#         x = self.head(x)
        return x

    def get_intermediate_layers(self, x):
        x = self.patch_embed(x)
        batch_size, seq_len, _ = x.size()

        cls_tokens = self.cls_token.expand(batch_size, -1, -1)  # stole cls_tokens impl from Phil Wang, thanks
        x = torch.cat((cls_tokens, x), dim=1)
        if self.pos_embed is not None:
            x = x + self.pos_embed
        x = self.pos_drop(x)

        features = []
        rel_pos_bias = self.rel_pos_bias() if self.rel_pos_bias is not None else None
        for blk in self.blocks:
            x = blk(x, rel_pos_bias)
            features.append(x)

        return features
    
    
def interpolate_pos_embed(model, checkpoint_model):
    if 'pos_embed' in checkpoint_model:
        pos_embed_checkpoint = checkpoint_model['pos_embed'].float()
        embedding_size = pos_embed_checkpoint.shape[-1]
        num_patches = model.patch_embed.num_patches
        num_extra_tokens = model.pos_embed.shape[-2] - num_patches
        # height (== width) for the checkpoint position embedding
        orig_size = int((pos_embed_checkpoint.shape[-2] - num_extra_tokens) ** 0.5)
        # height (== width) for the new position embedding
        new_size = int(num_patches ** 0.5)
        # class_token and dist_token are kept unchanged
        if orig_size != new_size:
            print("Position interpolate from %dx%d to %dx%d" % (orig_size, orig_size, new_size, new_size))
            extra_tokens = pos_embed_checkpoint[:, :num_extra_tokens]
            # only the position tokens are interpolated
            pos_tokens = pos_embed_checkpoint[:, num_extra_tokens:]
            pos_tokens = pos_tokens.reshape(-1, orig_size, orig_size, embedding_size).permute(0, 3, 1, 2)
            pos_tokens = torch.nn.functional.interpolate(
                pos_tokens, size=(new_size, new_size), mode='bicubic', align_corners=False)
            pos_tokens = pos_tokens.permute(0, 2, 3, 1).flatten(1, 2)
            new_pos_embed = torch.cat((extra_tokens, pos_tokens), dim=1)
            checkpoint_model['pos_embed'] = new_pos_embed
            
            
def convert_weights_to_fp16(model: nn.Module):
    """Convert applicable model parameters to fp16"""

    def _convert_weights_to_fp16(l):
        if isinstance(l, (nn.Conv1d, nn.Conv2d, nn.Linear)):
            l.weight.data = l.weight.data.half()
            if l.bias is not None:
                l.bias.data = l.bias.data.half()

#         if isinstance(l, (nn.MultiheadAttention, Attention)):
#             for attr in [*[f"{s}_proj_weight" for s in ["in", "q", "k", "v"]], "in_proj_bias", "bias_k", "bias_v"]:
#                 tensor = getattr(l, attr)
#                 if tensor is not None:
#                     tensor.data = tensor.data.half()

    model.apply(_convert_weights_to_fp16)
    
    
def create_eva_vit_g(img_size=224,drop_path_rate=0.4,use_checkpoint=False,precision="fp16"):
    model = VisionTransformer(
        img_size=img_size,
        patch_size=14,
        use_mean_pooling=False,
        embed_dim=1408,
        depth=39,
        num_heads=1408//88,
        mlp_ratio=4.3637,
        qkv_bias=True,
        drop_path_rate=drop_path_rate,
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        use_checkpoint=use_checkpoint,
    )  
    url = "https://storage.googleapis.com/sfr-vision-language-research/LAVIS/models/BLIP2/eva_vit_g.pth"
    cached_file = download_cached_file(
        url, check_hash=False, progress=True
    )
    state_dict = torch.load(cached_file, map_location="cpu")    
    interpolate_pos_embed(model,state_dict)
    
    incompatible_keys = model.load_state_dict(state_dict, strict=False)
#     print(incompatible_keys)
    
    if precision == "fp16":
#         model.to("cuda") 
        convert_weights_to_fp16(model)
    return model


"""DisCoVQA — Temporal Distortion-Content Transformers for VQA.

IEEE 2023 — separates temporal distortion extraction from
content-aware temporal attention using transformers.

GitHub: https://github.com/VQAssessment/DisCoVQA

discovqa_score — higher = better quality
"""

import logging
import cv2
import numpy as np
from typing import Optional

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class DisCoVQAModule(PipelineModule):
    name = "discovqa"
    description = "DisCoVQA temporal distortion-content VQA (2023)"
    default_config = {
        "subsample": 8,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 8)
        self._model = None
        self._backend = "heuristic"

    def setup(self) -> None:
        try:
            import discovqa
            self._model = discovqa
            self._backend = "native"
            logger.info("DisCoVQA (native) initialised")
            return
        except ImportError:
            pass

        self._backend = "heuristic"
        logger.info("DisCoVQA (heuristic) — install discovqa for full model")

    def process(self, sample: Sample) -> Sample:
        try:
            score = (
                float(self._model.predict(str(sample.path)))
                if self._backend == "native"
                else self._process_heuristic(sample)
            )

            if score is not None:
                if sample.quality_metrics is None:
                    sample.quality_metrics = QualityMetrics()
                sample.quality_metrics.discovqa_score = score

        except Exception as e:
            logger.warning(f"DisCoVQA failed for {sample.path}: {e}")

        return sample

    def _process_heuristic(self, sample: Sample) -> Optional[float]:
        """Heuristic: distortion extraction + content-aware temporal attention."""
        frames = self._extract_frames(sample)
        if not frames:
            return None

        # Distortion features per frame
        distortion_scores = []
        content_features = []

        for frame in frames:
            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)

            # Distortion: blur + noise + compression artifacts
            lap_var = cv2.Laplacian(gray, cv2.CV_64F).var()
            blur_dist = 1.0 - min(lap_var / 500.0, 1.0)  # Higher = more distorted

            noise_var = cv2.Sobel(gray, cv2.CV_64F, 1, 1).var()
            noise_dist = min(noise_var * 0.0001, 1.0)

            distortion = 0.6 * blur_dist + 0.4 * noise_dist
            distortion_scores.append(distortion)

            # Content: edge density as complexity proxy
            edges = cv2.Canny(frame, 50, 150)
            complexity = np.mean(edges > 0)
            content_features.append(complexity)

        # Content-aware temporal attention
        content_arr = np.array(content_features)
        weights = content_arr / (content_arr.sum() + 1e-8)  # Attend more to complex frames

        # Temporal distortion variation
        dist_arr = np.array(distortion_scores)
        weighted_distortion = np.sum(weights * dist_arr)
        temporal_var = np.var(dist_arr)

        # Quality = 1 - distortion, penalise temporal inconsistency
        quality = 1.0 - weighted_distortion
        temporal_penalty = min(temporal_var * 2.0, 0.3)
        score = quality - temporal_penalty

        return float(np.clip(score, 0.0, 1.0))

    def _extract_frames(self, sample: Sample):
        frames = []
        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            if total <= 0:
                cap.release()
                return []
            indices = np.linspace(0, total - 1, min(self.subsample, total), dtype=int)
            for idx in indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret, frame = cap.read()
                if ret:
                    frames.append(frame)
            cap.release()
        else:
            img = cv2.imread(str(sample.path))
            if img is not None:
                frames.append(img)
        return frames



from typing import Dict, List
import random
import re
from PIL import Image

from .utils import sample_video, read_image

class MultiImagesParser:
    def __init__(
        self,
        n_frames=8,
        is_training=True,
    ):
        self.n_frames = n_frames
        self.is_training = is_training
        # fmt: off
        self.data_temp = {
            "text": [
                [{
                    "prompt": "Describe the image in short.",
                    "response": "A rollerblader rides high in a full pipe while others watch"
                }],
                [{
                    "prompt": "Describe the image in short.",
                    "response": "A woman in winter clothes is on the sidewalk with a phone."
                }]
            ],
            "image": [
                {
                    "image_file": "/mnt/bn/videonaslq/images/flickr30k/images/3371533654.jpg"
                },
                {
                    "image_file": "/mnt/bn/videonaslq/images/coco/train2014/COCO_train2014_000000177950.jpg"
                },
                {
                    "video_file": "/mnt/bn/llmdatalq/jiangnan/video_generation/webvid_10M_download/20230609/videos/011851_011900/1047443473.mp4",
                    "frame_indices": [0, 85, 171, 256, 342, 427, 513, 598]
                }
            ],
            "dataset": "coco",
            "task": "multi_images",
            "image_processing_config": {},
        }
        # fmt: on
    
    def check_format(self, data_dict: Dict, image_processing_config: Dict):
        assert data_dict['dataset'] in ['coco', 'sharegpt4v_cap100k', 'sharegpt4v_mix665k', 'webvid', 'movie'], data_dict

        # 目前多图数据应该没有包含坐标的数据吧
        if image_processing_config.get('has_coordinates', False):
            raise ValueError(f'do_crop and has_coordinates cannot be True at the same time in MultiImagesParser!')
        
        # 检查是否能匹配到坐标
        texts = data_dict['text']
        for text in texts:
            match = re.search(r'\[(\d+(\.\d+)?,\s*)+\d+(\.\d+)?\]', text['prompt'] + text['response'])
            if match:
                print(f'[Warning] 疑似检测到包含坐标的数据：{data_dict}')

    
    def transform(self, data_dict: Dict, image_processing_config: Dict = None) -> Dict:
        self.check_format(data_dict, image_processing_config)

        # shuffle
        texts = data_dict['text']
        images = data_dict['image']
        images = self.load_images(images)
        idxs = list(range(len(texts)))
        random.shuffle(idxs)
        texts = [texts[i] for i in idxs]
        images = [images[i] for i in idxs]

        # sample n_frames
        if isinstance(self.n_frames, int):
            n_frames = random.choice(list(range(1, self.n_frames + 1)))
        else:
            n_frames = random.choice(self.n_frames)
        texts = texts[: n_frames]
        images = images[: n_frames]

        dataset = data_dict['dataset']
        if dataset in ['coco', 'sharegpt4v_cap100k', 'webvid', 'movie']:
            prompt, response = self.transform_for_caption_task(texts, dataset, images)
        else:
            prompt, response = self.transform_for_qa_task(texts, dataset, images)

        messages = [
            {
                "role": "user",
                "content": [
                    *[{"type": "image", "image": img} for img in images],
                    {"type": "text", "text": prompt},
                ]
            },
            {
                "role": "assistant",
                "content": [
                    {"type": "text", "text": response}
                ]
            }
        ]

        return messages
    
    def transform_for_caption_task(self, texts, dataset, images):
        idx = random.choice(list(range(len(texts))))

        if dataset == 'coco':
            if len(texts) == 1:
                prompt = 'Describe the image in short.'
            else:
                prompt = f'Describe the images starting from frame {idx + 1} in short in order.'
        elif dataset == 'sharegpt4v_cap100k':
            if len(texts) == 1:
                prompt = 'Describe the image in detail.'
            else:
                prompt = f'Describe the images starting from frame {idx + 1} in detail in order.'
        else:
            if len(texts) == 1:
                prompt = 'Describe the image.'
            else:
                prompt = f'Describe the images starting from frame {idx + 1} in order.'
        response = ''
        for i, text in enumerate(texts):
            if i < idx:
                continue
            if not isinstance(text, dict):
                text = random.choice(text)
            resp = text['response']
            response += f'{resp}\n'
        return prompt, response
    
    def transform_for_qa_task(self, texts, dataset, images):
        prompt, response = '', ''
        for i, text in enumerate(texts):
            if not isinstance(text, dict):
                text = random.choice(text)
            if len(texts) > 1:
                prompt += f'Question for frame {i+1}:\n' + text['prompt'] + '\n'
                response += f'Answer to question of frame {i+1}:\n' + text['response'] + '\n'
            else:
                prompt += text['prompt'] + '\n'
                response += text['response'] + '\n'
        return prompt, response


    def load_images(self, image_items: List[Dict]) -> List[Image.Image]:
        """
        image_items: List[Dict]. each item like:
            {"video_file": "path/to/video", "frame_indices": [1]}
            or
            {"image_file": "path/to/image"}    
        """
        if image_items is None:
            raise ValueError(f'image_items is None!')

        if isinstance(image_items, dict):
            image_items = [image_items]

        images = []

        for image_item in image_items:

            if 'video_file' in image_item:
                file_key = 'video_file'
            elif 'image_file' in image_item:
                file_key = 'image_file'
            else:
                raise KeyError(f'video_file or image_file not in {image_item}')

            file_path = image_item[file_key]
            if file_key == 'video_file':
                frame_indices = image_item.get('frame_indices', None)
                if frame_indices is None:
                    raise ValueError(f'read 0 frame: {image_item}')
                if isinstance(frame_indices, int):
                    frame_indices = [frame_indices]
                frames = sample_video(file_path, frame_indices = frame_indices)
                images.extend(frames)
            else:
                if isinstance(file_path, str):
                    file_path = [file_path]
                images.extend([read_image(f) for f in file_path])

        return images

if __name__ == '__main__':
    # python3 -m xenon_generation.data.custom_data_parsers.multi_images_parser

    from tqdm import tqdm
    from tools.rw_utils import read_jsonlines

    lines = read_jsonlines('/mnt/bn/videonaslq/VideoCaption/datasets_1009/sharegpt4v_cap100k/part_36.jsonl')
    lines = lines[:10]
    parser = MultiImagesParser(n_frames=8)
    for i, l in tqdm(enumerate(lines)):
        l_image_processing_config = l.get('image_processing_config', {})
        messages = parser.transform(l, l_image_processing_config)
        print(messages)


import os
import random

import cv2
import decord
import pandas as pd
import torch
from PIL import Image
from torch.utils.data import Dataset


class VideoRetrievalDataset(Dataset):
    def __init__(
        self,
        csv_path,
        dataset_dir,
        preprocessor,
        video_ext="mp4",
        num_frames=8,
        multi_sent=False,
    ):
        self.data = pd.read_csv(csv_path)
        self.dataset_dir = dataset_dir
        self.video_ext = video_ext

        self.preprocessor = preprocessor
        self.num_frames = num_frames
        self.multi_sent = multi_sent

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        video_id = self.data["video_id"].values[index]
        sentences = self.data["sentence"].values[index]
        if self.multi_sent:
            sentences = sentences.split("@")
        else:
            sentences = [sentences]
        video_path = os.path.join(
            self.dataset_dir, "{}.{}".format(video_id, self.video_ext)
        )

        images = self._load_video(video_path)

        images = [
            (
                self.preprocessor(image.convert("RGB"))
                if image.mode == "L"
                else self.preprocessor(image)
            )
            for image in images
        ]

        return images, sentences

    def _load_video(self, media_path):
        vr = decord.VideoReader(media_path)
        total_frames = len(vr)
        if self.num_frames == 1:
            frame_indices = [total_frames // 2]
        else:
            frame_indices = [
                int(i * (total_frames - 1) / (self.num_frames - 1))
                for i in range(self.num_frames)
            ]
        try:
            images = vr.get_batch(frame_indices).asnumpy()
        except Exception as e:
            cap = cv2.VideoCapture(media_path)
            images = []
            for pos in frame_indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, pos)
                ret, frame = cap.read()
                if ret:
                    # Convert the frame from BGR to RGB
                    rgb_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                    images.append(rgb_frame)
                else:
                    break

        images = [Image.fromarray(image) for image in images]

        return images



import logging
from pathlib import Path
import urllib.request

import cv2
import numpy as np
from ayase.models import Sample, ValidationIssue, ValidationSeverity, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)

# Original URLs:
#   https://github.com/VQAssessment/FAST-VQA-and-FasterVQA/releases/download/v2.0.0/FAST_VQA_3D_1_1.pth
#   https://github.com/VQAssessment/FAST-VQA-and-FasterVQA/releases/download/v2.0.0/FAST_VQA_B_1_4.pth
#   https://github.com/VQAssessment/FAST-VQA-and-FasterVQA/releases/download/v2.0.0/FAST_VQA_M_1_4.pth
FASTVQA_MODEL_URLS = {
    "FAST_VQA_3D_1_1.pth": "https://huggingface.co/AkaneTendo25/ayase-models/resolve/main/fast_vqa/FAST_VQA_3D_1_1.pth",
    "FAST_VQA_B_1_4.pth": "https://huggingface.co/AkaneTendo25/ayase-models/resolve/main/fast_vqa/FAST_VQA_B_1_4.pth",
    "FAST_VQA_M_1_4.pth": "https://huggingface.co/AkaneTendo25/ayase-models/resolve/main/fast_vqa/FAST_VQA_M_1_4.pth",
}


def resize_frame_maintain_aspect(frame, min_dim):
    if min_dim <= 0:
        return frame
    h, w = frame.shape[:2]
    if h <= min_dim and w <= min_dim:
        return frame
    if h < w:
        scale = min_dim / h
    else:
        scale = min_dim / w
    if scale >= 1.0:
        return frame
    new_h = int(h * scale)
    new_w = int(w * scale)
    return cv2.resize(frame, (new_w, new_h), interpolation=cv2.INTER_AREA)


class FastVQAModule(PipelineModule):
    name = "fast_vqa"
    description = "Deep Learning Video Quality Assessment (FAST-VQA)"
    default_config = {"model_type": "FasterVQA"}

    def __init__(self, config=None):
        super().__init__(config)
        self.model_type = self.config.get(
            "model_type", "FasterVQA"
        )  # FasterVQA, FasterVQA-MS, FasterVQA-MT, FAST-VQA, FAST-VQA-M
        self.device = "cpu"
        self._model = None
        self._opt = None
        self._ml_available = False

        self.fastvqa_mean_std = {
            "FasterVQA": (0.14759505, 0.03613452),
            "FasterVQA-MS": (0.15218826, 0.03230298),
            "FasterVQA-MT": (0.14699507, 0.036453716),
            "FAST-VQA": (-0.110198185, 0.04178565),
            "FAST-VQA-M": (0.023889644, 0.030781006),
        }

    def setup(self) -> None:
        try:
            import torch
            import decord
            from ayase.third_party.fastvqa.models import DiViDeAddEvaluator
            from ayase.third_party.fastvqa.datasets import get_spatial_fragments

            self.device = "cuda" if torch.cuda.is_available() else "cpu"
            logger.info(f"Setting up FastVQA ({self.model_type}) on {self.device}...")

            # Load options
            opts_map = {
                "FasterVQA": "f3dvqa-b.yml",
                "FasterVQA-MS": "fastervqa-ms.yml",
                "FasterVQA-MT": "fastervqa-mt.yml",
                "FAST-VQA": "fast-b.yml",
                "FAST-VQA-M": "fast-m.yml",
            }

            if self.model_type not in opts_map:
                logger.warning(f"Unknown model type {self.model_type}, defaulting to FasterVQA")
                self.model_type = "FasterVQA"

            opt_file = opts_map[self.model_type]
            # Assume options are in src/ayase/fastvqa/options
            # We need to find the absolute path
            import ayase.third_party.fastvqa

            base_path = Path(ayase.third_party.fastvqa.__file__).parent
            opt_path = base_path / "options" / opt_file

            if not opt_path.exists():
                raise FileNotFoundError(f"Config file not found: {opt_path}")

            import yaml  # type: ignore[import-untyped]
            with open(opt_path, "r") as f:
                self._opt = yaml.safe_load(f)

            # Initialize model
            self._model = DiViDeAddEvaluator(**self._opt["model"]["args"]).to(self.device)

            # Download/Load weights
            models_dir = Path(self.config.get("models_dir") or "models")
            weights_root = models_dir / "fast_vqa"
            shared_downloads = models_dir / "downloads"
            weights_filename = str(self._opt["test_load_path"]).replace("*", "_").split("/")[-1]
            weights_path = weights_root / weights_filename
            legacy_path = base_path / "pretrained_weights" / weights_filename
            shared_path = shared_downloads / weights_filename

            if not weights_path.exists():
                if legacy_path.exists():
                    weights_root.mkdir(parents=True, exist_ok=True)
                    legacy_path.replace(weights_path)
                elif shared_path.exists():
                    weights_root.mkdir(parents=True, exist_ok=True)
                    shared_path.replace(weights_path)
                else:
                    self._download_weights(weights_filename, weights_path)

            checkpoint = torch.load(str(weights_path), map_location=self.device, weights_only=True)
            if "state_dict" in checkpoint:
                self._model.load_state_dict(checkpoint["state_dict"])
            else:
                self._model.load_state_dict(checkpoint)

            self._model.eval()
            self._ml_available = True

        except ImportError:
            logger.warning("Missing dependencies (decord, timm, or yaml). FastVQA disabled.")
        except Exception as e:
            logger.warning(f"Failed to setup FastVQA: {e}")
            import traceback

            traceback.print_exc()

    def _download_weights(self, filename, destination_path):
        url = FASTVQA_MODEL_URLS.get(filename)
        if not url:
            # Fallback map for filenames that might not match exactly
            if "FAST_VQA_3D" in filename:
                url = FASTVQA_MODEL_URLS["FAST_VQA_3D_1_1.pth"]
            elif "FAST_VQA_B" in filename:
                url = FASTVQA_MODEL_URLS["FAST_VQA_B_1_4.pth"]
            elif "FAST_VQA_M" in filename:
                url = FASTVQA_MODEL_URLS["FAST_VQA_M_1_4.pth"]
            else:
                raise ValueError(f"Unknown weight file: {filename}")

        destination_path.parent.mkdir(parents=True, exist_ok=True)
        logger.info(f"Downloading {filename} from {url}...")

        urllib.request.urlretrieve(url, destination_path)
        logger.info(f"Downloaded {filename}")

    def _prepare_input(self, video_path, min_dimension=None):
        import torch
        import decord
        from ayase.third_party.fastvqa.datasets import get_spatial_fragments, FragmentSampleFrames, SampleFrames

        # Logic adapted from quality_scorer.py
        val_config = None
        for key in self._opt["data"].keys():
            if key.startswith("val-"):
                val_config = self._opt["data"][key]["args"]
                break
        if val_config is None:
            val_config = self._opt["data"]["train"]["args"]

        t_data_opt = val_config
        s_data_opt = t_data_opt["sample_types"]

        vsamples = {}
        clip_count = None

        video_reader = decord.VideoReader(str(video_path))
        total_frames = len(video_reader)

        for sample_type, sample_args in s_data_opt.items():
            frame_interval = sample_args.get("frame_interval", t_data_opt.get("frame_interval", 1))

            if t_data_opt.get("t_frag", 1) > 1:
                sampler = FragmentSampleFrames(
                    fsize_t=sample_args["clip_len"] // sample_args.get("t_frag", 1),
                    fragments_t=sample_args.get("t_frag", 1),
                    num_clips=sample_args.get("num_clips", 1),
                    frame_interval=frame_interval,
                )
            else:
                sampler = SampleFrames(
                    clip_len=sample_args["clip_len"],
                    frame_interval=frame_interval,
                    num_clips=sample_args.get("num_clips", 1),
                )

            frames = sampler(total_frames)
            frame_dict = {idx: video_reader[idx] for idx in np.unique(frames)}

            imgs = []
            for idx in frames:
                frame_obj = frame_dict[idx]
                if hasattr(frame_obj, "asnumpy"):
                    frame = frame_obj.asnumpy()
                elif hasattr(frame_obj, "numpy"):
                    frame = frame_obj.numpy()
                else:
                    frame = np.array(frame_obj)

                # H, W, C, BGR (decord default is RGB if not specified? wait, decord is RGB usually)
                # Decord returns RGB by default.
                # Wait, quality_scorer says: frame = frame_dict[idx].asnumpy() # H, W, C, BGR format from decord
                # BUT decord documentation says it returns RGB.
                # Let's assume RGB for now as Ayase uses RGB.

                # quality_scorer code: frame_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                # This implies decord gave BGR. Let's verify.
                # Actually standard decord is RGB. Maybe they configured it differently or it's a specific version.
                # Ayase uses RGB everywhere.

                # Let's assume decord returns RGB.
                frame_rgb = frame

                if min_dimension and min_dimension > 0:
                    frame_rgb = resize_frame_maintain_aspect(frame_rgb, min_dimension)

                imgs.append(torch.from_numpy(frame_rgb))

            video = torch.stack(imgs, 0).permute(3, 0, 1, 2)  # C, T, H, W

            sampled_video = get_spatial_fragments(video, **sample_args)

            mean = torch.FloatTensor([123.675, 116.28, 103.53])
            std = torch.FloatTensor([58.395, 57.12, 57.375])
            sampled_video = ((sampled_video.permute(1, 2, 3, 0) - mean) / std).permute(3, 0, 1, 2)

            num_clips = sample_args.get("num_clips", 1)
            if clip_count is None:
                clip_count = num_clips

            sampled_video = sampled_video.reshape(
                sampled_video.shape[0], num_clips, -1, *sampled_video.shape[2:]
            ).transpose(0, 1)
            vsamples[sample_type] = sampled_video.contiguous()

        return {"samples": vsamples, "num_clips": clip_count}

    def process(self, sample: Sample) -> Sample:
        if not self._ml_available or not sample.is_video:
            return sample

        try:
            import torch

            prepared = self._prepare_input(sample.path)

            batched_inputs = {}
            for sample_type in prepared["samples"].keys():
                batched_inputs[sample_type] = prepared["samples"][sample_type].to(self.device)

            with torch.no_grad():
                scores = self._model(batched_inputs)
                # scores shape: (num_clips, 1) usually
                # We average over clips
                raw_score = scores.mean().item()

            # Sigmoid rescale
            mean, std = self.fastvqa_mean_std[self.model_type]
            x = (raw_score - mean) / std
            final_score = 1 / (1 + np.exp(-x))
            final_score_100 = final_score * 100.0

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()

            sample.quality_metrics.fast_vqa_score = final_score_100

            if final_score_100 < 40.0:
                sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"Low FastVQA Score: {final_score_100:.1f}",
                        details={"score": final_score_100, "raw": raw_score},
                    )
                )

        except Exception as e:
            logger.warning(f"FastVQA failed for {sample.path}: {e}")
            import traceback

            traceback.print_exc()

        return sample



# Copyright (2024) Bytedance Ltd. and/or its affiliates

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ..models.modeling_tarsier import TarsierForConditionalGeneration, LlavaConfig
# from dataset.processor import Processor
from ..dataset.tarsier_datamodule import init_processor
import torch
import base64
from ..tools.color import Color
import yaml
import sys
def load_model_and_processor(model_name_or_path, data_config):
    print(Color.red(f"Load model and processor from: {model_name_or_path}"), flush=True)
    if isinstance(data_config, str):
        data_config = yaml.safe_load(open(data_config, 'r'))
    processor = init_processor(model_name_or_path, data_config)
    model_config = LlavaConfig.from_pretrained(
        model_name_or_path,
        trust_remote_code=True,
    )
    model = TarsierForConditionalGeneration.from_pretrained(
        model_name_or_path,
        config=model_config,
        device_map='auto',
        torch_dtype=torch.bfloat16,
        trust_remote_code=True
    )
    model.eval()
    return model, processor

def file_to_base64(img_path):
    with open(img_path, 'rb') as video_file:
        video_b64_str = base64.b64encode(video_file.read()).decode()
    return video_b64_str




"""Tests for dataset_analytics module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_dataset_analytics_basics():
    from ayase.modules.dataset_analytics import DatasetAnalyticsModule
    _test_module_basics(DatasetAnalyticsModule, "dataset_analytics")

def test_dataset_analytics_extract(video_sample):
    from ayase.modules.dataset_analytics import DatasetAnalyticsModule
    m = DatasetAnalyticsModule()
    feat = m.extract_features(video_sample)
    # May be None for non-video or missing deps
    assert video_sample is not None



"""Tests for identity_loss module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_identity_loss_basics():
    from ayase.modules.identity_loss import IdentityLossModule
    _test_module_basics(IdentityLossModule, "identity_loss")

def test_identity_loss_image(image_sample):
    from ayase.modules.identity_loss import IdentityLossModule
    image_sample.quality_metrics = QualityMetrics()
    m = IdentityLossModule()
    m.on_mount()
    result = m.process(image_sample)
    assert result is image_sample

def test_identity_loss_video(video_sample):
    from ayase.modules.identity_loss import IdentityLossModule
    video_sample.quality_metrics = QualityMetrics()
    m = IdentityLossModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



from functools import partial

from detectron2 import model_zoo
from detectron2.config import LazyCall as L
from detectron2.modeling.backbone.vit import get_vit_lr_decay_rate
from detectron2.solver import WarmupParamScheduler
from fvcore.common.param_scheduler import MultiStepParamScheduler

from ..common.coco_loader_lsj import dataloader

model = model_zoo.get_config("common/models/mask_rcnn_vitdet.py").model

# Initialization and trainer settings
train = model_zoo.get_config("common/train.py").train
train.amp.enabled = True
train.ddp.fp16_compression = True
train.init_checkpoint = "detectron2://ImageNetPretrained/MAE/mae_pretrain_vit_base.pth?matching_heuristics=True"


# Schedule
# 100 ep = 184375 iters * 64 images/iter / 118000 images/ep
train.max_iter = 184375

lr_multiplier = L(WarmupParamScheduler)(
    scheduler=L(MultiStepParamScheduler)(
        values=[1.0, 0.1, 0.01],
        milestones=[163889, 177546],
        num_updates=train.max_iter,
    ),
    warmup_length=250 / train.max_iter,
    warmup_factor=0.001,
)

# Optimizer
optimizer = model_zoo.get_config("common/optim.py").AdamW
optimizer.params.lr_factor_func = partial(
    get_vit_lr_decay_rate, num_layers=12, lr_decay_rate=0.7
)
optimizer.params.overrides = {"pos_embed": {"weight_decay": 0.0}}



"""Tests for physics module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_physics_basics():
    from ayase.modules.physics import PhysicsModule
    _test_module_basics(PhysicsModule, "physics")

def test_physics_video(video_sample):
    from ayase.modules.physics import PhysicsModule
    video_sample.quality_metrics = QualityMetrics()
    m = PhysicsModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""Tests for scene_detection module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_scene_detection_basics():
    from ayase.modules.scene_detection import SceneDetectionModule
    _test_module_basics(SceneDetectionModule, "scene_detection")

def test_scene_detection_video(video_sample):
    from ayase.modules.scene_detection import SceneDetectionModule
    video_sample.quality_metrics = QualityMetrics()
    m = SceneDetectionModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""FUNQUE (Fused Unified Quality Evaluator) module.

Full-reference quality metric that fuses SSIM, VIF, and DLM features.

Backend tiers:
  1. **funque package** — real FUNQUE from ``pip install funque``
     (``github.com/abhinaukumar/funque``)
  2. **Handcrafted FR** — OpenCV-based SSIM/VIF/DLM approximation (FR mode)
  3. **Handcrafted NR** — self-referencing proxy (NR mode, no reference needed)
"""

import logging
from pathlib import Path
from typing import Optional

import numpy as np

from ayase.models import QualityMetrics, Sample
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class FUNQUEModule(PipelineModule):
    name = "funque"
    description = "Fused quality evaluator (FUNQUE package, handcrafted FR, or NR fallback)"
    default_config = {"subsample": 8}

    def __init__(self, config: Optional[dict] = None) -> None:
        super().__init__(config)
        self._funque_available = False
        self._backend = "heuristic_nr"

    def setup(self) -> None:
        # Tier 1: Real FUNQUE package
        try:
            import funque as _funque
            self._funque_available = True
            self._backend = "funque"
            logger.info("FUNQUE loaded real funque package")
            return
        except (ImportError, Exception) as e:
            logger.info("funque package unavailable: %s", e)

        # Tiers 2/3 are always available (OpenCV-based)
        self._backend = "heuristic_fr"
        logger.info("FUNQUE using handcrafted approximation")

    def process(self, sample: Sample) -> Sample:
        if sample.quality_metrics is None:
            sample.quality_metrics = QualityMetrics()

        reference_path = getattr(sample, "reference_path", None)
        has_reference = reference_path is not None and Path(str(reference_path)).exists()

        try:
            if self._funque_available and has_reference:
                self._process_funque(sample, Path(str(reference_path)))
            elif has_reference:
                self._process_fr_heuristic(sample, Path(str(reference_path)))
            else:
                self._process_nr_heuristic(sample)
        except Exception as e:
            logger.warning("FUNQUE failed: %s", e)
        return sample

    def _process_funque(self, sample: Sample, reference_path: Path) -> None:
        """Process using the real FUNQUE package."""
        import funque

        try:
            # FUNQUE API: compute quality score between reference and distorted
            result = funque.compute(
                str(reference_path),
                str(sample.path),
            )
            if isinstance(result, dict) and "funque" in result:
                score = float(result["funque"])
            elif isinstance(result, (int, float)):
                score = float(result)
            else:
                score = float(result)

            sample.quality_metrics.funque_score = float(np.clip(score, 0.0, 1.0))
        except Exception as e:
            logger.info("FUNQUE package failed: %s, falling back to heuristic", e)
            self._process_fr_heuristic(sample, reference_path)

    def _process_fr_heuristic(self, sample: Sample, reference_path: Path) -> None:
        """Full-reference heuristic: compare sample against reference."""
        import cv2

        frames_dist = self._load_frames(sample)
        if not frames_dist:
            return

        # Load reference frames
        ref_sample_path = reference_path
        if ref_sample_path.suffix.lower() in (".mp4", ".avi", ".mov", ".mkv", ".webm"):
            frames_ref = self._load_video_frames(ref_sample_path)
        else:
            frame = cv2.imread(str(ref_sample_path))
            frames_ref = [frame] if frame is not None else []

        if not frames_ref:
            self._process_nr_heuristic(sample)
            return

        # Compare paired frames
        n_pairs = min(len(frames_dist), len(frames_ref))
        frame_scores = []
        for i in range(n_pairs):
            ref_gray = cv2.cvtColor(frames_ref[i], cv2.COLOR_BGR2GRAY).astype(np.float64)
            dist_gray = cv2.cvtColor(frames_dist[i], cv2.COLOR_BGR2GRAY).astype(np.float64)

            # Resize to match
            h, w = dist_gray.shape
            ref_gray = cv2.resize(ref_gray, (w, h))

            score = self._compute_fr_features(ref_gray, dist_gray)
            frame_scores.append(score)

        # Temporal pooling
        if len(frame_scores) > 1:
            sorted_scores = sorted(frame_scores)
            n = len(sorted_scores)
            p5 = sorted_scores[max(0, int(n * 0.05))]
            final = 0.4 * p5 + 0.4 * np.mean(frame_scores) + 0.2 * np.median(frame_scores)
        else:
            final = frame_scores[0]

        sample.quality_metrics.funque_score = float(np.clip(final, 0.0, 1.0))

    def _compute_fr_features(self, ref: np.ndarray, dist: np.ndarray) -> float:
        """Compute FR quality features between reference and distorted frames."""
        import cv2

        # 1. SSIM approximation
        C1 = (0.01 * 255) ** 2
        C2 = (0.03 * 255) ** 2
        mu_ref = cv2.GaussianBlur(ref, (11, 11), 1.5)
        mu_dist = cv2.GaussianBlur(dist, (11, 11), 1.5)
        sigma_ref_sq = cv2.GaussianBlur(ref ** 2, (11, 11), 1.5) - mu_ref ** 2
        sigma_dist_sq = cv2.GaussianBlur(dist ** 2, (11, 11), 1.5) - mu_dist ** 2
        sigma_ref_dist = cv2.GaussianBlur(ref * dist, (11, 11), 1.5) - mu_ref * mu_dist

        ssim_map = ((2 * mu_ref * mu_dist + C1) * (2 * sigma_ref_dist + C2)) / \
                   ((mu_ref ** 2 + mu_dist ** 2 + C1) * (sigma_ref_sq + sigma_dist_sq + C2))
        ssim_score = float(np.mean(ssim_map))

        # 2. VIF approximation (information fidelity)
        eps = 1e-10
        sigma_ref_sq = np.maximum(sigma_ref_sq, 0)
        g = sigma_ref_dist / (sigma_ref_sq + eps)
        sigma_v = sigma_dist_sq - g * sigma_ref_dist
        sigma_v = np.maximum(sigma_v, eps)
        sigma_n = 2.0  # Noise variance estimate
        vif_num = np.log2(1 + g ** 2 * sigma_ref_sq / (sigma_v + sigma_n))
        vif_den = np.log2(1 + sigma_ref_sq / sigma_n)
        vif_score = float(np.sum(vif_num) / (np.sum(vif_den) + eps))
        vif_score = min(1.0, max(0.0, vif_score))

        # 3. DLM (Detail Loss Measure) approximation
        hp_ref = ref - cv2.GaussianBlur(ref, (15, 15), 3.0)
        hp_dist = dist - cv2.GaussianBlur(dist, (15, 15), 3.0)
        detail_loss = np.mean(np.abs(hp_ref - hp_dist))
        dlm_score = max(0.0, 1.0 - detail_loss / 50.0)

        return 0.35 * ssim_score + 0.35 * vif_score + 0.30 * dlm_score

    def _process_nr_heuristic(self, sample: Sample) -> None:
        """No-reference heuristic (self-referencing)."""
        import cv2

        frames = self._load_frames(sample)
        if not frames:
            return

        frame_scores = []
        for frame in frames:
            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)
            score = self._compute_nr_features(gray)
            frame_scores.append(score)

        if len(frame_scores) > 1:
            sorted_scores = sorted(frame_scores)
            n = len(sorted_scores)
            p5 = sorted_scores[max(0, int(n * 0.05))]
            final = 0.4 * p5 + 0.4 * np.mean(frame_scores) + 0.2 * np.median(frame_scores)
        else:
            final = frame_scores[0]

        sample.quality_metrics.funque_score = float(np.clip(final, 0.0, 1.0))

    def _compute_nr_features(self, gray: np.ndarray) -> float:
        """Compute NR quality features from a grayscale frame."""
        import cv2

        h, w = gray.shape

        # Multi-scale SSIM-like luminance/contrast
        ssim_features = []
        current = gray.copy()
        for scale in range(3):
            mu = cv2.GaussianBlur(current, (11, 11), 1.5)
            mu_sq = mu ** 2
            sigma_sq = cv2.GaussianBlur(current ** 2, (11, 11), 1.5) - mu_sq
            sigma_sq = np.maximum(sigma_sq, 0)
            lum = np.mean(mu) / 255.0
            con = np.mean(np.sqrt(sigma_sq + 1e-8)) / 128.0
            ssim_features.extend([lum, min(1.0, con)])
            current = cv2.resize(current, (max(1, w // (2 ** (scale + 1))),
                                           max(1, h // (2 ** (scale + 1)))))

        # VIF-like information content
        vif_features = []
        current = gray.copy()
        for scale in range(3):
            mu = cv2.GaussianBlur(current, (5, 5), 1.0)
            var = cv2.GaussianBlur((current - mu) ** 2, (5, 5), 1.0)
            info = np.log2(1.0 + np.mean(var))
            vif_features.append(min(1.0, info / 12.0))
            current = cv2.resize(current, (max(1, current.shape[1] // 2),
                                           max(1, current.shape[0] // 2)))

        # DLM-like detail energy
        hp = gray - cv2.GaussianBlur(gray, (15, 15), 3.0)
        detail_energy = np.mean(hp ** 2)
        detail_score = min(1.0, detail_energy / 500.0)

        # Gradient structural features
        gx = cv2.Sobel(gray, cv2.CV_64F, 1, 0, ksize=3)
        gy = cv2.Sobel(gray, cv2.CV_64F, 0, 1, ksize=3)
        grad_mag = np.sqrt(gx ** 2 + gy ** 2)
        grad_coherence = np.std(grad_mag) / (np.mean(grad_mag) + 1e-8)
        structural = min(1.0, 1.0 / (1.0 + abs(grad_coherence - 1.0)))

        ssim_score = np.mean(ssim_features)
        vif_score = np.mean(vif_features)

        return float(0.30 * ssim_score + 0.25 * vif_score + 0.25 * detail_score + 0.20 * structural)

    def _load_frames(self, sample: Sample) -> list:
        import cv2

        subsample = self.config.get("subsample", 8)
        frames = []
        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            indices = list(range(0, total, max(1, total // subsample)))[:subsample]
            for idx in indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret, frame = cap.read()
                if ret:
                    frames.append(frame)
            cap.release()
        else:
            frame = cv2.imread(str(sample.path))
            if frame is not None:
                frames.append(frame)
        return frames

    def _load_video_frames(self, path: Path) -> list:
        import cv2

        subsample = self.config.get("subsample", 8)
        frames = []
        cap = cv2.VideoCapture(str(path))
        total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        indices = list(range(0, total, max(1, total // subsample)))[:subsample]
        for idx in indices:
            cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
            ret, frame = cap.read()
            if ret:
                frames.append(frame)
        cap.release()
        return frames



"""TIFA — Text-to-Image Faithfulness Assessment (ICCV 2023).

Measures how faithfully a generated image matches its text prompt by:
    1. Generating yes/no questions from the caption (rule-based, no LLM)
    2. Answering them via VQA on the generated image
    3. Scoring = fraction of correct answers

Requires ``sample.caption.text`` — skips if no caption is available.
Also checks for a sidecar ``.txt`` file next to the sample.

Tiered backends:
    1. Rule-based question generation + ViLT VQA (dandelin/vilt-b32-finetuned-vqa)
    2. CLIP text-image cosine similarity proxy
    3. Heuristic — caption complexity-based estimate
"""

import logging
import re
from pathlib import Path
from typing import List, Optional, Tuple

import cv2
import numpy as np

from ayase.models import CaptionMetadata, QualityMetrics, Sample
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)

# Common colors for question generation
_COLORS = {
    "red", "blue", "green", "yellow", "orange", "purple", "pink",
    "black", "white", "brown", "gray", "grey", "golden", "silver",
}

# Words to skip as nouns (stop words / too generic)
_SKIP_NOUNS = {
    "a", "an", "the", "it", "this", "that", "there", "here",
    "image", "photo", "picture", "scene", "view", "background",
    "something", "thing", "one", "ones", "way", "lot", "kind",
}


def _generate_questions(caption: str, max_questions: int = 8) -> List[Tuple[str, str]]:
    """Generate (question, expected_answer) pairs from a caption.

    Returns simple yes/no questions about objects, colors, and counts
    mentioned in the caption text.  Pure rule-based — no LLM needed.
    """
    questions: List[Tuple[str, str]] = []
    caption_lower = caption.lower()
    words = re.findall(r"[a-z]+", caption_lower)

    # --- Color questions ---
    found_colors = [c for c in _COLORS if c in words]
    for color in found_colors[:2]:
        questions.append((f"Is there something {color} in the image?", "yes"))

    # --- Noun/object questions ---
    # Simple POS-free heuristic: take nouns as words after determiners/adjectives
    # or standalone content words that are likely nouns.
    _NON_NOUNS = {
        "is", "are", "was", "were", "has", "have", "had", "be",
        "with", "from", "into", "onto", "upon", "over", "under",
        "and", "but", "for", "nor", "yet", "not", "can", "will",
        "very", "much", "more", "most", "also", "just", "only",
    }
    _ADJECTIVE_SUFFIXES = ("ing", "ly", "ed", "ful", "ous", "ive", "ish", "able", "ible", "less")
    _COMMON_ADJECTIVES = {
        "big", "small", "large", "tall", "short", "long", "new", "old",
        "good", "bad", "high", "low", "dark", "bright", "beautiful",
        "nice", "pretty", "ugly", "happy", "sad", "fast", "slow",
        "soft", "hard", "warm", "cold", "hot", "cool", "young",
    }

    def _is_likely_noun(w: str) -> bool:
        """Return True if the word is likely a noun (not a verb/adjective)."""
        if w in _NON_NOUNS or w in _COMMON_ADJECTIVES:
            return False
        if w.endswith(_ADJECTIVE_SUFFIXES):
            return False
        return True

    candidate_nouns = []
    for i, w in enumerate(words):
        if w in _SKIP_NOUNS or len(w) < 3:
            continue
        if w in _COLORS:
            continue
        # Words following a/an/the are likely nouns
        if i > 0 and words[i - 1] in ("a", "an", "the"):
            if _is_likely_noun(w):
                candidate_nouns.append(w)
        # Standalone content words that pass the noun filter
        elif _is_likely_noun(w):
            candidate_nouns.append(w)

    seen = set()
    for noun in candidate_nouns:
        if noun in seen:
            continue
        seen.add(noun)
        questions.append((f"Is there a {noun} in the image?", "yes"))
        if len(questions) >= max_questions:
            break

    # --- Count questions ---
    count_pattern = re.findall(r"(\d+|two|three|four|five|six|seven|eight|nine|ten)\s+(\w+)", caption_lower)
    for num, obj in count_pattern[:2]:
        questions.append((f"Are there {num} {obj} in the image?", "yes"))
        if len(questions) >= max_questions:
            break

    return questions[:max_questions]


class TIFAModule(PipelineModule):
    name = "tifa"
    description = "TIFA text-to-image faithfulness via VQA question answering (ICCV 2023)"
    default_config = {
        "vqa_model": "dandelin/vilt-b32-finetuned-vqa",
        "num_questions": 8,
        "subsample": 4,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.vqa_model = self.config.get("vqa_model", "dandelin/vilt-b32-finetuned-vqa")
        self.num_questions = self.config.get("num_questions", 8)
        self.subsample = self.config.get("subsample", 4)
        self._backend = None  # "vilt" | "clip" | "heuristic"
        self._model = None
        self._processor = None
        self._clip_model = None
        self._clip_processor = None
        self._device = "cpu"

    def setup(self):
        # Tier 1: ViLT VQA
        try:
            import torch
            from transformers import ViltForQuestionAnswering, ViltProcessor

            self._device = "cuda" if torch.cuda.is_available() else "cpu"
            models_dir = self.config.get("models_dir", "models")
            self._processor = ViltProcessor.from_pretrained(self.vqa_model, cache_dir=models_dir)
            self._model = ViltForQuestionAnswering.from_pretrained(
                self.vqa_model, cache_dir=models_dir
            ).to(self._device)
            self._backend = "vilt"
            logger.info(f"TIFA: using ViLT VQA backend on {self._device}.")
            return
        except Exception:
            pass

        # Tier 2: CLIP proxy
        try:
            import torch
            from transformers import CLIPModel, CLIPProcessor

            self._device = "cuda" if torch.cuda.is_available() else "cpu"
            models_dir = self.config.get("models_dir", "models")
            clip_name = "openai/clip-vit-base-patch32"
            self._clip_model = CLIPModel.from_pretrained(clip_name, cache_dir=models_dir).to(
                self._device
            )
            self._clip_processor = CLIPProcessor.from_pretrained(clip_name, cache_dir=models_dir)
            self._backend = "clip"
            logger.info("TIFA: using CLIP similarity proxy.")
            return
        except Exception:
            pass

        # Tier 3: Heuristic
        self._backend = "heuristic"
        logger.info("TIFA: using heuristic fallback (no ML models).")

    def process(self, sample: Sample) -> Sample:
        caption_text = self._get_caption(sample)
        if not caption_text:
            return sample

        try:
            if self._backend == "vilt":
                score = self._compute_vilt(sample, caption_text)
            elif self._backend == "clip":
                score = self._compute_clip(sample, caption_text)
            else:
                score = self._compute_heuristic(caption_text)

            if score is None:
                return sample

            score = float(np.clip(score, 0.0, 1.0))

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()
            sample.quality_metrics.tifa_score = score

        except Exception as e:
            logger.warning(f"TIFA failed for {sample.path}: {e}")

        return sample

    # -- Caption extraction -----------------------------------------------------

    def _get_caption(self, sample: Sample) -> Optional[str]:
        if sample.caption and sample.caption.text:
            return sample.caption.text

        # Check sidecar .txt file
        txt_path = sample.path.with_suffix(".txt")
        if txt_path.exists():
            text = txt_path.read_text(encoding="utf-8").strip()
            if text:
                return text

        return None

    # -- Backend implementations ------------------------------------------------

    def _compute_vilt(self, sample: Sample, caption: str) -> Optional[float]:
        from PIL import Image

        frames = self._load_frames(sample)
        if not frames:
            return None

        questions = _generate_questions(caption, self.num_questions)
        if not questions:
            return None

        correct_total = 0
        count_total = 0

        for frame in frames:
            pil_img = Image.fromarray(frame)
            for question, expected in questions:
                try:
                    encoding = self._processor(pil_img, question, return_tensors="pt").to(
                        self._device
                    )
                    outputs = self._model(**encoding)
                    idx = outputs.logits.argmax(-1).item()
                    answer = self._model.config.id2label[idx].lower()
                    if answer == expected:
                        correct_total += 1
                    count_total += 1
                except Exception:
                    continue

        if count_total == 0:
            return None
        return correct_total / count_total

    def _compute_clip(self, sample: Sample, caption: str) -> Optional[float]:
        import torch
        from PIL import Image

        frames = self._load_frames(sample)
        if not frames:
            return None

        similarities = []
        for frame in frames:
            pil_img = Image.fromarray(frame)
            inputs = self._clip_processor(
                text=[caption], images=pil_img, return_tensors="pt", padding=True
            ).to(self._device)
            with torch.no_grad():
                outputs = self._clip_model(**inputs)
            # Cosine similarity from CLIP logits
            sim = outputs.logits_per_image.item() / 100.0  # normalize to ~0-1
            similarities.append(float(np.clip(sim, 0.0, 1.0)))

        return float(np.mean(similarities)) if similarities else None

    def _compute_heuristic(self, caption: str) -> float:
        """Estimate TIFA from caption complexity: shorter/simpler = easier to match."""
        words = caption.split()
        n = len(words)
        if n <= 5:
            return 0.75
        elif n <= 15:
            return 0.60
        elif n <= 30:
            return 0.45
        else:
            return 0.35

    # -- Frame loading ----------------------------------------------------------

    def _load_frames(self, sample: Sample):
        frames = []
        try:
            if sample.is_video:
                cap = cv2.VideoCapture(str(sample.path))
                total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                if total <= 0:
                    cap.release()
                    return frames
                n = min(self.subsample, total)
                indices = np.linspace(0, total - 1, n, dtype=int)
                for idx in indices:
                    cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                    ret, frame = cap.read()
                    if ret:
                        frames.append(cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
                cap.release()
            else:
                img = cv2.imread(str(sample.path))
                if img is not None:
                    frames.append(cv2.cvtColor(img, cv2.COLOR_BGR2RGB))
        except Exception as e:
            logger.debug(f"Frame loading failed: {e}")
        return frames



# Copyright (c) Meta Platforms, Inc. and affiliates.

import json
import logging
import os
import re
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional, Tuple

import torch
import torch.distributed as dist
import torch.distributed.checkpoint as dcp
import torch.nn as nn
import torch.optim.optimizer
from omegaconf import OmegaConf
from torch.distributed._tensor import DeviceMesh
from torch.distributed.checkpoint import FileSystemReader
from torch.distributed.checkpoint.format_utils import dcp_to_torch_save
from torch.distributed.checkpoint.state_dict import (get_model_state_dict,
                                                     get_state_dict,
                                                     set_state_dict)

from ..core.distributed import get_is_master

logger = logging.getLogger("CHECKPOINT")

FOLDER_NAME = "{:010d}"
RE_FOLDER = r"\d{10}"

RE_CKPT = r"__\d_\d\.distcp"

CONSOLIDATE_FOLDER = "consolidated"
CONSOLIDATE_NAME = "consolidated.pth"

CONFIG_NAME = "params.json"
TRAIN_STATE_NAME = "train_state_{:05d}.json"
RE_DIGITS = re.compile(r"\d+")


@dataclass
class SaveEvery:
    every: int = 1000
    keep: int = 0


@dataclass
class CheckpointArgs:
    dump: SaveEvery = field(default_factory=SaveEvery)
    eval: SaveEvery = field(default_factory=SaveEvery)
    path: Optional[str] = None
    init_ckpt_path: Optional[str] = None
    vision_model_path: Optional[str] = None
    is_consolidated_model: bool = False
    continue_training_from_init: bool = False


def _get_key_step(name: str):
    return int(re.findall(RE_DIGITS, name)[-1])


def consolidate_checkpoints(ckpt_dir: str):
    """
    Consolidates all FSDP checkpoints in a directory to a single file
    Consolidate checkpoint is saved in a subdirectory of ckpt_dir

    Parameters:
        ckpt_dir: str - path to the directory containing the checkpoints

    Returns the path to the consolidated checkpoint
    """
    consolidate_path = Path(ckpt_dir) / CONSOLIDATE_FOLDER
    if not (consolidate_path / CONSOLIDATE_NAME).exists():
        consolidate_path.mkdir(exist_ok=True)
        logger.info(f"Consolidating to: {str(consolidate_path)}")
        dcp_to_torch_save(ckpt_dir, str(consolidate_path / CONSOLIDATE_NAME))
        (consolidate_path / CONFIG_NAME).write_text(
            (Path(ckpt_dir) / CONFIG_NAME).read_text()
        )
        logger.info("Consolidated !")
    return consolidate_path


def load_from_checkpoint(
    ckpt_dir: str,
    model: nn.Module,
    optimizer: Optional[torch.optim.Optimizer] = None,
    model_key: str = "model",
    optim_key: str = "optim",
):
    if not (Path(ckpt_dir) / ".metadata").exists():
        raise ValueError(
            f"Please convert the checkpoint distcp format using `torch.distributed.checkpoint.format_utils.torch_save_to_dcp` before loading it"
        )

    state_dict = {}
    if optimizer is not None:
        state_dict[model_key], state_dict[optim_key] = get_state_dict(model, optimizer)
    else:
        state_dict[model_key] = get_model_state_dict(model)
        if model_key == "":  # If only loading a model directly, the key should be empty
            state_dict = state_dict.pop(model_key)

    dcp.load(state_dict, checkpoint_id=ckpt_dir)


class CheckpointManager:
    def __init__(self, args: CheckpointArgs):
        self.path = args.path
        self.dump_every = args.dump
        self.eval_every = args.eval
        self.init_ckpt_path = args.init_ckpt_path
        self.continue_training_from_init = args.continue_training_from_init

        assert os.path.exists(
            self.path
        ), f"Path {self.path} does not exist and needs to be created before using CheckpointManager (use instantiate_and_make_dir)"

        self.existing_saves = self.get_existing_saves()

    def get_existing_saves(self) -> List[Path]:
        folders = [
            p
            for p in Path(self.path).iterdir()
            if p.is_dir() and re.match(RE_FOLDER, p.name)
        ]
        folders.sort(key=lambda p: _get_key_step(p.name))
        return folders

    def clean_up(self):
        logger.info("Cleaning up checkpoints...")
        dump_folders = []
        eval_folders = []
        other_folders = []
        for p in self.existing_saves:
            is_dump = _get_key_step(p.name) % self.dump_every.every == 0
            is_eval = _get_key_step(p.name) % self.eval_every.every == 0
            if is_dump:
                dump_folders.append(p)
            if is_eval:
                eval_folders.append(p)
            if not (is_dump or is_eval):
                other_folders.append(p)

        logger.info(f"Dump folders: {dump_folders}")
        logger.info(f"Eval folders: {eval_folders}")
        logger.info(f"Other folders: {other_folders}")

        if self.dump_every.keep > 0:
            dump_folders = dump_folders[-self.dump_every.keep :]
        if self.eval_every.keep > 0:
            eval_folders = eval_folders[-self.eval_every.keep :]

        folder_to_keep = set(other_folders + dump_folders + eval_folders)
        folder_to_remove = set(self.existing_saves) - folder_to_keep

        logger.info(f"Removing folders: {folder_to_remove}")

        if dist.get_rank() == 0:
            for folder in folder_to_remove:
                for file in folder.iterdir():
                    if file.is_file():
                        file.unlink()
                    elif file.is_dir():
                        assert file.name in [CONSOLIDATE_FOLDER]
                        for f in file.iterdir():
                            f.unlink()
                        file.rmdir()
                folder.rmdir()

        dist.barrier()

        self.existing_saves = list(folder_to_keep)
        self.existing_saves.sort(key=lambda p: _get_key_step(p.name))

    def get_last_step_path(self, dp_rank: int = 0) -> Optional[Path]:
        path = None
        for p in reversed(self.existing_saves):
            if (p / TRAIN_STATE_NAME.format(dp_rank)).is_file():
                path = p
                break
        return path

    def _create_folder(self, base_path: Path, folder_name: str) -> Path:
        folder = base_path / folder_name
        if get_is_master():
            folder.mkdir(parents=False, exist_ok=True)
        if dist.is_initialized():
            dist.barrier()
        return folder

    def _get_dp_tp_mesh(
        self, device_mesh: Optional[DeviceMesh] = None
    ) -> Tuple[int, int]:
        dp_rank = 0
        tp_rank = 0
        if device_mesh is not None:
            if "dp_replicate" in device_mesh.mesh_dim_names:
                dp_rank = device_mesh.get_local_rank("dp_replicate")
                if "dp_shard" in device_mesh.mesh_dim_names:
                    dp_rank = dp_rank * device_mesh[
                        "dp_replicate"
                    ].size() + device_mesh.get_local_rank("dp_shard")
            if "tp" in device_mesh.mesh_dim_names:
                tp_rank = device_mesh.get_local_rank("tp")
        return dp_rank, tp_rank

    @torch.no_grad()
    def get_state_dict(
        self,
        model,
        optimizer,
    ):
        model_sd, optim_sd = get_state_dict(model, optimizer)
        return {"model": model_sd, "optim": optim_sd}

    def save(
        self,
        model,
        optimizer,
        train_state,
        config,
        device_mesh: Optional[DeviceMesh] = None,
    ) -> bool:

        # When creating directory check if only rank0 or is there other solution
        path = Path(self.path)
        curr_save_dir = self._create_folder(path, FOLDER_NAME.format(train_state.step))
        logger.info(f"Saving to: {str(curr_save_dir)}")

        if dist.is_initialized():
            dist.barrier()

        logger.info("Saving...")
        state_dict = self.get_state_dict(model, optimizer)
        dcp.save(state_dict, checkpoint_id=curr_save_dir)
        logger.info("State dict saved!")

        if dist.is_initialized():
            dist.barrier()

        if get_is_master():
            with open(curr_save_dir / CONFIG_NAME, "w") as f:
                json.dump(
                    OmegaConf.to_container(OmegaConf.structured(config), resolve=True),
                    f,
                    indent=4,
                )

        # Add json dump here
        dp_rank, tp_rank = self._get_dp_tp_mesh(device_mesh)
        if tp_rank == 0:
            train_state_name = TRAIN_STATE_NAME.format(dp_rank)
            logger.info(
                f"Saving train state to: {str(curr_save_dir / train_state_name)}"
            )
            # logger.info(f"train_state.state_dict()={train_state.state_dict()}")
            with open(curr_save_dir / train_state_name, "w") as f:
                json.dump(train_state.state_dict(), f)
            logger.info("Train state saved !")

        self.existing_saves.append(curr_save_dir)

        self.clean_up()

        if dist.is_initialized():
            dist.barrier()
        return True

    @torch.no_grad()
    def load(
        self,
        model: nn.Module,
        optimizer,
        train_state,
        device_mesh: DeviceMesh,
        path: Optional[Path] = None,
    ):
        dp_rank, tp_rank = self._get_dp_tp_mesh(device_mesh)
        # Loading tries to load the provided path, if not available the last saved step and finally from the init path
        path = path or self.get_last_step_path(dp_rank=dp_rank)
        # If none of those are available don't do anything
        if path is None:
            # If no checkpoints exist do nothing
            return

        # Only load train state if it's provided, the files exist and we're not loading from init path
        train_state_name = TRAIN_STATE_NAME.format(dp_rank)
        logger.info("Reloading train state")
        with open(path / train_state_name, "r") as f:
            train_state_dict = json.load(f)
        train_state.load_state_dict(train_state_dict)
        logger.info("Train state reloaded")

        logger.info(f"Loading from: {str(path)}")
        state_dict = self.get_state_dict(
            model=model,
            optimizer=optimizer,
        )
        dcp.load(state_dict, checkpoint_id=path)
        logger.info("State dict loaded.")

        logger.info("Reloading model and optim")

        set_state_dict(
            model,
            optimizer,
            model_state_dict=state_dict["model"],
            optim_state_dict=state_dict["optim"],
        )
        logger.info("Model and optim reloaded")

    @classmethod
    def instantiate_and_make_dir(cls, args: CheckpointArgs):
        if get_is_master():
            os.makedirs(args.path, exist_ok=True)
        dist.barrier()

        return cls(args)


def get_consolidated_ckpt_path(ckpt_dir: Path, mp_rank: int = 0, mp_size: int = 1):
    if mp_size == 1:
        assert mp_rank == 0
        no_rank_path = ckpt_dir / "consolidated.pth"
        if no_rank_path.exists():
            return no_rank_path
    return ckpt_dir / f"consolidated.{mp_rank:02d}.pth"


def load_consolidated_checkpoint(
    model: nn.Module,
    consolidated_path: str,
    vision_model_path: Optional[str] = None,
):
    """
    Loads a consolidated checkpoint into the model.
    This version supports both:
      - a single file named 'consolidated.pth'
      - multiple parts named like 'consolidated.00.pth', 'consolidated.01.pth', etc.
    """
    ckpt_path = Path(consolidated_path)
    cp_file = get_consolidated_ckpt_path(ckpt_path, mp_rank=0, mp_size=1)
    if cp_file.exists():
        # Use the single file
        st_dict = torch.load(cp_file, weights_only=True)
        if "model" in st_dict:
            st_dict = st_dict["model"]
    else:
        # Fall back to multi-part consolidated files (e.g. consolidated.00.pth, consolidated.01.pth, …)
        checkpoint_files = sorted(ckpt_path.glob("consolidated.*.pth"))
        if not checkpoint_files:
            raise FileNotFoundError(
                f"No consolidated checkpoint file found in {ckpt_path}."
            )
        st_dict = {}
        for ckpt_file in checkpoint_files:
            part = torch.load(ckpt_file, weights_only=True)
            # If the checkpoint part is wrapped with "model", unwrap it
            if "model" in part:
                part = part["model"]
            # Merge the state dicts (assumes the keys are all unique or will correctly overwrite)
            st_dict.update(part)

    model.vision_projector.init_tensors()
    model.vision_model.init_tensors()
    model.rope_embeddings.reset_parameters()

    if vision_model_path is not None:
        model.vision_model.load_ckpt(vision_model_path)

    missing_keys, unexpected_keys = model.load_state_dict(st_dict, strict=False)
    missing_keys = [k for k in missing_keys if "tied_module.weight" not in k]
    if vision_model_path is not None:
        # vision_model is already loaded separately
        missing_keys = [k for k in missing_keys if "vision_model." not in k]
    if len(missing_keys) > 0:
        logger.warning(f"Missing keys when reloading: {missing_keys}")
    if len(unexpected_keys) > 0:
        logger.warning(f"Unexpected keys when reloading: {unexpected_keys}")



"""
 Copyright (c) 2022, salesforce.com, inc.
 All rights reserved.
 SPDX-License-Identifier: BSD-3-Clause
 For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
"""

import logging
import os

import torch
from ...common.dist_utils import download_cached_file
from ...common.utils import is_url
from ...models.base_model import BaseModel
from ...models.vit import interpolate_pos_embed
from transformers import BertTokenizer


class BlipBase(BaseModel):
    @classmethod
    def init_tokenizer(cls):
        tokenizer = BertTokenizer.from_pretrained("bert-base-uncased")
        tokenizer.add_special_tokens({"bos_token": "[DEC]"})
        tokenizer.add_special_tokens({"additional_special_tokens": ["[ENC]"]})
        tokenizer.enc_token_id = tokenizer.additional_special_tokens_ids[0]
        return tokenizer

    def load_from_pretrained(self, url_or_filename):
        if is_url(url_or_filename):
            cached_file = download_cached_file(
                url_or_filename, check_hash=False, progress=True
            )
            checkpoint = torch.load(cached_file, map_location="cpu")
        elif os.path.isfile(url_or_filename):
            checkpoint = torch.load(url_or_filename, map_location="cpu")
        else:
            raise RuntimeError("checkpoint url or path is invalid")

        state_dict = checkpoint["model"]

        state_dict["visual_encoder.pos_embed"] = interpolate_pos_embed(
            state_dict["visual_encoder.pos_embed"], self.visual_encoder
        )
        if "visual_encoder_m.pos_embed" in self.state_dict().keys():
            state_dict["visual_encoder_m.pos_embed"] = interpolate_pos_embed(
                state_dict["visual_encoder_m.pos_embed"], self.visual_encoder_m
            )

        for key in self.state_dict().keys():
            if key in state_dict.keys():
                if state_dict[key].shape != self.state_dict()[key].shape:
                    del state_dict[key]

        msg = self.load_state_dict(state_dict, strict=False)

        logging.info("Missing keys {}".format(msg.missing_keys))
        logging.info("load checkpoint from %s" % url_or_filename)

        return msg



import hashlib
import os
import urllib
import warnings
from packaging import version
from typing import Union, List

import torch
from PIL import Image
from torchvision.transforms import Compose, Resize, CenterCrop, ToTensor, Normalize
from tqdm import tqdm

from .model import build_model
from .simple_tokenizer import SimpleTokenizer as _Tokenizer

try:
    from torchvision.transforms import InterpolationMode
    BICUBIC = InterpolationMode.BICUBIC
except ImportError:
    BICUBIC = Image.BICUBIC


if version.parse(torch.__version__) < version.parse("1.7.1"):
    warnings.warn("PyTorch version 1.7.1 or higher is recommended")


__all__ = ["available_models", "load", "tokenize"]
_tokenizer = _Tokenizer()

_MODELS = {
    "RN50": "https://openaipublic.azureedge.net/clip/models/afeb0e10f9e5a86da6080e35cf09123aca3b358a0c3e3b6c78a7b63bc04b6762/RN50.pt",
    "RN101": "https://openaipublic.azureedge.net/clip/models/8fa8567bab74a42d41c5915025a8e4538c3bdbe8804a470a72f30b0d94fab599/RN101.pt",
    "RN50x4": "https://openaipublic.azureedge.net/clip/models/7e526bd135e493cef0776de27d5f42653e6b4c8bf9e0f653bb11773263205fdd/RN50x4.pt",
    "RN50x16": "https://openaipublic.azureedge.net/clip/models/52378b407f34354e150460fe41077663dd5b39c54cd0bfd2b27167a4a06ec9aa/RN50x16.pt",
    "RN50x64": "https://openaipublic.azureedge.net/clip/models/be1cfb55d75a9666199fb2206c106743da0f6468c9d327f3e0d0a543a9919d9c/RN50x64.pt",
    "ViT-B/32": "https://openaipublic.azureedge.net/clip/models/40d365715913c9da98579312b702a82c18be219cc2a73407c4526f58eba950af/ViT-B-32.pt",
    "ViT-B/16": "https://openaipublic.azureedge.net/clip/models/5806e77cd80f8b59890b7e101eabd078d9fb84e6937f9e85e4ecb61988df416f/ViT-B-16.pt",
    "ViT-L/14": "https://openaipublic.azureedge.net/clip/models/b8cca3fd41ae0c99ba7e8951adf17d267cdb84cd88be6f7c2e0eca1737a03836/ViT-L-14.pt",
    "ViT-L/14@336px": "https://openaipublic.azureedge.net/clip/models/3035c92b350959924f9f00213499208652fc7ea050643e8b385c2dac08641f02/ViT-L-14-336px.pt",
}

_ENV_MODEL_PATHS = {
    "ViT-B/32": "AYASE_BRIGHTVQ_CLIP_VIT_B32_PATH",
    "ViT-L/14": "AYASE_BRIGHTVQ_CLIP_VIT_L14_PATH",
}


def _download(url: str, root: str):
    os.makedirs(root, exist_ok=True)
    filename = os.path.basename(url)

    expected_sha256 = url.split("/")[-2]
    download_target = os.path.join(root, filename)

    if os.path.exists(download_target) and not os.path.isfile(download_target):
        raise RuntimeError(f"{download_target} exists and is not a regular file")

    if os.path.isfile(download_target):
        if hashlib.sha256(open(download_target, "rb").read()).hexdigest() == expected_sha256:
            return download_target
        else:
            warnings.warn(f"{download_target} exists, but the SHA256 checksum does not match; re-downloading the file")

    with urllib.request.urlopen(url) as source, open(download_target, "wb") as output:
        with tqdm(total=int(source.info().get("Content-Length")), ncols=80, unit='iB', unit_scale=True, unit_divisor=1024) as loop:
            while True:
                buffer = source.read(8192)
                if not buffer:
                    break

                output.write(buffer)
                loop.update(len(buffer))

    if hashlib.sha256(open(download_target, "rb").read()).hexdigest() != expected_sha256:
        raise RuntimeError("Model has been downloaded but the SHA256 checksum does not not match")

    return download_target


def _convert_image_to_rgb(image):
    return image.convert("RGB")


def _transform(n_px):
    return Compose([
        Resize(n_px, interpolation=BICUBIC),
        CenterCrop(n_px),
        _convert_image_to_rgb,
        ToTensor(),
        Normalize((0.48145466, 0.4578275, 0.40821073), (0.26862954, 0.26130258, 0.27577711)),
    ])


def available_models() -> List[str]:
    """Returns the names of available CLIP models"""
    return list(_MODELS.keys())


def load(name: str, device: Union[str, torch.device] = "cuda" if torch.cuda.is_available() else "cpu", jit: bool = False, download_root: str = None):
    """Load a CLIP model

    Parameters
    ----------
    name : str
        A model name listed by `clip.available_models()`, or the path to a model checkpoint containing the state_dict

    device : Union[str, torch.device]
        The device to put the loaded model

    jit : bool
        Whether to load the optimized JIT model or more hackable non-JIT model (default).

    download_root: str
        path to download the model files; by default, it uses "~/.cache/clip"

    Returns
    -------
    model : torch.nn.Module
        The CLIP model

    preprocess : Callable[[PIL.Image], torch.Tensor]
        A torchvision transform that converts a PIL image into a tensor that the returned model can take as its input
    """
    env_model_path = os.environ.get(_ENV_MODEL_PATHS.get(name, ""))
    if env_model_path and os.path.isfile(env_model_path):
        model_path = env_model_path
    elif name in _MODELS:
        model_path = _download(
            _MODELS[name],
            download_root
            or os.environ.get("AYASE_BRIGHTVQ_CLIP_DOWNLOAD_ROOT")
            or os.path.expanduser("~/.cache/clip"),
        )
    elif os.path.isfile(name):
        model_path = name
    else:
        raise RuntimeError(f"Model {name} not found; available models = {available_models()}")

    with open(model_path, 'rb') as opened_file:
        try:
            # loading JIT archive
            model = torch.jit.load(opened_file, map_location=device if jit else "cpu").eval()
            state_dict = None
        except RuntimeError:
            # loading saved state dict
            if jit:
                warnings.warn(f"File {model_path} is not a JIT archive. Loading as a state dict instead")
                jit = False
            state_dict = torch.load(opened_file, map_location="cpu")

    if not jit:
        model = build_model(state_dict or model.state_dict()).to(device)
        if str(device) == "cpu":
            model.float()
        return model, _transform(model.visual.input_resolution)

    # patch the device names
    device_holder = torch.jit.trace(lambda: torch.ones([]).to(torch.device(device)), example_inputs=[])
    device_node = [n for n in device_holder.graph.findAllNodes("prim::Constant") if "Device" in repr(n)][-1]

    def _node_get(node: torch._C.Node, key: str):
        """Gets attributes of a node which is polymorphic over return type.
        
        From https://github.com/pytorch/pytorch/pull/82628
        """
        sel = node.kindOf(key)
        return getattr(node, sel)(key)

    def patch_device(module):
        try:
            graphs = [module.graph] if hasattr(module, "graph") else []
        except RuntimeError:
            graphs = []

        if hasattr(module, "forward1"):
            graphs.append(module.forward1.graph)

        for graph in graphs:
            for node in graph.findAllNodes("prim::Constant"):
                if "value" in node.attributeNames() and str(_node_get(node, "value")).startswith("cuda"):
                    node.copyAttributes(device_node)

    model.apply(patch_device)
    patch_device(model.encode_image)
    patch_device(model.encode_text)

    # patch dtype to float32 on CPU
    if str(device) == "cpu":
        float_holder = torch.jit.trace(lambda: torch.ones([]).float(), example_inputs=[])
        float_input = list(float_holder.graph.findNode("aten::to").inputs())[1]
        float_node = float_input.node()

        def patch_float(module):
            try:
                graphs = [module.graph] if hasattr(module, "graph") else []
            except RuntimeError:
                graphs = []

            if hasattr(module, "forward1"):
                graphs.append(module.forward1.graph)

            for graph in graphs:
                for node in graph.findAllNodes("aten::to"):
                    inputs = list(node.inputs())
                    for i in [1, 2]:  # dtype can be the second or third argument to aten::to()
                        if _node_get(inputs[i].node(), "value") == 5:
                            inputs[i].node().copyAttributes(float_node)

        model.apply(patch_float)
        patch_float(model.encode_image)
        patch_float(model.encode_text)

        model.float()

    return model, _transform(model.input_resolution.item())


def tokenize(texts: Union[str, List[str]], context_length: int = 77, truncate: bool = False) -> Union[torch.IntTensor, torch.LongTensor]:
    """
    Returns the tokenized representation of given input string(s)

    Parameters
    ----------
    texts : Union[str, List[str]]
        An input string or a list of input strings to tokenize

    context_length : int
        The context length to use; all CLIP models use 77 as the context length

    truncate: bool
        Whether to truncate the text in case its encoding is longer than the context length

    Returns
    -------
    A two-dimensional tensor containing the resulting tokens, shape = [number of input strings, context_length].
    We return LongTensor when torch version is <1.8.0, since older index_select requires indices to be long.
    """
    if isinstance(texts, str):
        texts = [texts]

    sot_token = _tokenizer.encoder["<|startoftext|>"]
    eot_token = _tokenizer.encoder["<|endoftext|>"]
    all_tokens = [[sot_token] + _tokenizer.encode(text) + [eot_token] for text in texts]
    if version.parse(torch.__version__) < version.parse("1.8.0"):
        result = torch.zeros(len(all_tokens), context_length, dtype=torch.long)
    else:
        result = torch.zeros(len(all_tokens), context_length, dtype=torch.int)

    for i, tokens in enumerate(all_tokens):
        if len(tokens) > context_length:
            if truncate:
                tokens = tokens[:context_length]
                tokens[-1] = eot_token
            else:
                raise RuntimeError(f"Input {texts[i]} is too long for context length {context_length}")
        result[i, :len(tokens)] = torch.tensor(tokens)

    return result
